import numpy as np
import SimpleITK as sitk
import scipy.io as sio


def makeSphere(shape, radius, position):
    # assume shape and position are both a 3-tuple of int or float
    # the units are pixels / voxels (px for short)
    # radius is a int or float in px
    semisizes = (radius,) * 3

    # genereate the grid for the support points
    # centered at the position indicated by position
    grid = [slice(-x0, dim - x0) for x0, dim in zip(position, shape)]
    position = np.ogrid[grid]
    # calculate the distance of all points from `position` center
    # scaled by the radius
    arr = np.zeros(shape, dtype=float)
    for x_i, semisize in zip(position, semisizes):
        # this can be generalized for exponent != 2
        # in which case `(x_i / semisize)`
        # would become `np.abs(x_i / semisize)`
        arr += (x_i / semisize) ** 2

    # the inner part of the sphere will have distance below 1
    result = (arr <= 1.0)
    result = result.astype(int)

    return result

def makeTransducer_vertical2(shape, step_size, ROC, width, tran_pose, target_pose):

    ROC = ROC/1000
    width = width/1000
    
    ## ROC depend on Voxel
    ROC = int(np.round(ROC / step_size))
    width = int(np.round(width / step_size))


    ## direction vector from target to geometrical focus
    dirVec = (target_pose-tran_pose)/np.linalg.norm(target_pose-tran_pose)
    geometry_center = target_pose+dirVec*np.abs(ROC-np.linalg.norm(target_pose-tran_pose))

    x_arr = np.arange(0,shape[0])
    y_arr = np.arange(0,shape[1])
    z_arr = np.arange(0,shape[2])

    x_idx = geometry_center[0]
    y_idx = geometry_center[1]
    z_idx = geometry_center[2]

    # Make mesh grid, order is VERY IMPORTANT!!
    #mx, my, mz = np.meshgrid(x_arr, y_arr, z_arr)

    my, mx, mz = np.meshgrid(y_arr, x_arr, z_arr)
    shape = mx.shape

    # Make large sphere for transducer
    sphere = makeSphere((shape[0], shape[1], shape[2]), radiusVN, (x_idx, y_idx, z_idx))
    sphere_sub = makeSphere((shape[0], shape[1], shape[2]), radiusVN-1, (x_idx, y_idx, z_idx))
    sphere = sphere - sphere_sub

    # Make plane to cut the sphere and make transducer
    # calculate normal vector of plane
    normal = (tran_pose-target_pose)/np.linalg.norm(tran_pose-target_pose)
    temp_length = np.sqrt(ROC**2 - (width/2)**2)

    # Find one point on the cutting plane
    plane_point = geometry_center + temp_length*normal

    # Plane equation using normal and point
    cont = normal[0]*plane_point[0] + normal[1]*plane_point[1] + normal[2]*plane_point[2]
    plane_cal = mx * normal[0] + my * normal[1] + mz * normal[2] - cont
    plane_cal = np.round(plane_cal, 1)

    plane_cal[plane_cal > 0] = 1
    plane_cal[plane_cal < 0] = 0
    plane_cal = plane_cal.astype(int)

    # Cut sphere using plane
    transducer = np.multiply(sphere, plane_cal)
    transducer = transducer.transpose([2,1,0])
    trans_idx = np.where(transducer==1)
    
    # Change order for sitk save
    sitk_transducer = transducer
    #sitk_transducer = sitk_transducer.transpose([2,1,0])
    spacing = np.array((step_size, step_size, step_size))
    origin = np.array((min(x_arr), min(y_arr), min(z_arr)))
    trans_itk = sitk.GetImageFromArray(sitk_transducer*2000, sitk.sitkInt8)
    trans_itk.SetSpacing(spacing)
    trans_itk.SetOrigin(origin)


    return transducer, trans_idx, trans_itk

def makeTransducer_vertical(image_cordi, step_size, ROC, width, tran_pose, target_pose):

    ## ROC depend on Voxel
    radiusVN = int(np.round(ROC / step_size))

    ## direction vector from target to geometrical focus
    dirVec = (target_pose-tran_pose)/np.linalg.norm(target_pose-tran_pose)
    geometry_center = target_pose+dirVec*np.abs(ROC-np.linalg.norm(target_pose-tran_pose))

    x_arr = image_cordi[0]
    y_arr = image_cordi[1]
    z_arr = image_cordi[2]

    x_idx = np.argmin(np.abs(x_arr-geometry_center[0]))
    y_idx = np.argmin(np.abs(y_arr-geometry_center[1]))
    z_idx = np.argmin(np.abs(z_arr-geometry_center[2]))

    # Make mesh grid, order is VERY IMPORTANT!!
    #mx, my, mz = np.meshgrid(x_arr, y_arr, z_arr)

    my, mx, mz = np.meshgrid(y_arr, x_arr, z_arr)
    shape = mx.shape

    # Make large sphere for transducer
    sphere = makeSphere((shape[0], shape[1], shape[2]), radiusVN, (x_idx, y_idx, z_idx))
    sphere_sub = makeSphere((shape[0], shape[1], shape[2]), radiusVN-1, (x_idx, y_idx, z_idx))
    sphere = sphere - sphere_sub

    # Make plane to cut the sphere and make transducer
    # calculate normal vector of plane
    normal = (tran_pose-target_pose)/np.linalg.norm(tran_pose-target_pose)
    temp_length = np.sqrt(ROC**2 - (width/2)**2)

    # Find one point on the cutting plane
    plane_point = geometry_center + temp_length*normal

    # Plane equation using normal and point
    cont = normal[0]*plane_point[0] + normal[1]*plane_point[1] + normal[2]*plane_point[2]
    plane_cal = mx * normal[0] + my * normal[1] + mz * normal[2] - cont
    plane_cal = np.round(plane_cal, 1)

    plane_cal[plane_cal > 0] = 1
    plane_cal[plane_cal < 0] = 0
    plane_cal = plane_cal.astype(int)

    # Cut sphere using plane
    transducer = np.multiply(sphere, plane_cal)
    transducer = transducer.transpose([2,1,0])

    # Change order for sitk save
    sitk_transducer = transducer
    #sitk_transducer = sitk_transducer.transpose([2,1,0])
    spacing = np.array((step_size, step_size, step_size))
    origin = np.array((min(x_arr), min(y_arr), min(z_arr)))
    trans_itk = sitk.GetImageFromArray(sitk_transducer*2000, sitk.sitkInt8)
    trans_itk.SetSpacing(spacing)
    trans_itk.SetOrigin(origin)


    return transducer, trans_itk

def makeBowlTransducer(Nx, Ny, Nz, step_size, ROC, width, PML = 10):

    Nx_half = int(Nx/2)
    Ny_half = int(Ny/2)

    cutting_len = np.sqrt(pow(ROC, 2) - pow(width/2, 2))
    cutting_lenVNM = int(np.round(cutting_len/step_size))

    radiusVN = int(np.round(ROC/step_size))
    radiusVNM = radiusVN + PML

    sphere = makeSphere((radiusVNM*4, radiusVNM*4, radiusVNM*2), radiusVN, (radiusVNM*2, radiusVNM*2, radiusVNM))
    sphere_sub  = makeSphere((radiusVNM*4, radiusVNM*4, radiusVNM*2), radiusVN-1, (radiusVNM*2, radiusVNM*2, radiusVNM))
    sphereN = sphere-sphere_sub

    # This function differ to the "makebowl" in k-wave, to match this difference x step margin was given
    function_difference_gap = 0
    radiusVNM2 = radiusVNM * 2
    Transducer = sphereN[radiusVNM2-Nx_half:radiusVNM2+Nx_half, radiusVNM2-Ny_half:radiusVNM2+Ny_half, function_difference_gap:(radiusVNM-cutting_lenVNM)+function_difference_gap]
    sensor_model = np.zeros((Nx,Ny,Nz))
    sensor_model[:,:,PML:(radiusVNM-cutting_lenVNM)+PML] = Transducer

    return sensor_model

def cordi2idx(image_array, point):

    idx = np.zeros((3))
    for i in range(3):
        idx[i] = np.abs(image_array[i] - point[i]).argmin()

    return idx

def make_transducer3(dx, Tcenter, Tnormal):

    Tdia = 0.065
    Troc = 0.071
    TP = 0

    H = Troc - np.sqrt(Troc**2 - (0.5 * Tdia)**2)
    TP_dis = np.floor(H / dx)

    nH = 3000
    nD = 3000
    nA = np.floor(0.5 * Tdia / dx) + 5
    A = np.zeros((2 * nA + 1, 2 * nA + 1, 2 * nA + 1))

    Vs = [0, - Tnormal[3], Tnormal[2]]
    Vs = Vs / np.linalg.norm(Vs)
    Vt = np.cross(Tnormal, Vs)

    pi = 3.141592653589793

    for i in range(nH):
        R = np.sqrt(Troc ^ 2 - (Troc - H + (i) * H / nH) ^ 2)
        for j in range(nD):
            theta = 2 * pi / nD * j
            X = (nA + 1) + np.round((-(i) * H / nH * Tnormal[1] + R * np.cos(theta) * Vs(1) + R * np.sin(theta) * Vt(1)) / dx)
            Y = (nA + 1) + np.round((-(i) * H / nH * Tnormal[2] + R * np.cos(theta) * Vs(2) + R * np.sin(theta) * Vt(2)) / dx)
            Z = (nA + 1) + np.round((-(i) * H / nH * Tnormal[3] + R * np.cos(theta) * Vs(3) + R * np.sin(theta) * Vt(3)) / dx)
            A[X,Y,Z] = 1

    nS = np.sum(A)
    Spos = np.zeros((nS, 4))
    count = 1;
    for i in range((2 * nA + 1)):
        for j in range((2 * nA + 1)):
            for k in range((2 * nA + 1)):
                if (A[i, j, k] == 1):
                    Spos[count, 1] = i - (nA + 1) + Tcenter[1];
                    Spos[count, 2] = j - (nA + 1) + Tcenter[2];
                    Spos[count, 3] = k - (nA + 1) + Tcenter[3];
                    Spos[count, 4] = TP * 2 * pi * (TP_dis - 1 + np.floor((i - nA + 1) * Tnormal[1] + (j - nA + 1) * Tnormal[2] + (k - nA + 1) * Tnormal[3])) / TP_dis;
                    count = count + 1;

def neper2db(alpha, y):

    alphaDB = 20*np.log10(np.exp(1))*alpha*pow((2*np.pi*1e6), y)/100

    return alphaDB

def saveITK(path, itkImage):
    writer = sitk.ImageFileWriter()
    writer.SetFileName(path)
    writer.Execute(itkImage)