
import vtk
import numpy as np
import numpy.matlib
import SimpleITK as sitk
import cv2
from scipy import ndimage
import matplotlib.pyplot as plt

l2n = lambda l: np.array(l)
n2l = lambda n: list(n)

from vtk.util import numpy_support as ns


class niiCook():

    def __init__(self):
        self.dummy = 0

    def readITK(self, image):

        spacing   = image.GetSpacing()
        origin    = image.GetOrigin()  ## bounds
        dimension = image.GetSize()
        extent = (0, dimension[0] - 1, 0, dimension[1] - 1, 0, dimension[2] - 1)

        array = sitk.GetArrayFromImage(image)

        self.itkImage = image
        self.spacing = spacing
        self.origin = origin
        self.dimension = dimension
        self.extent = extent
        self.array = array


    def readSavedFile(self, filePath):

        if filePath[-2:] == "gz":
            reader = sitk.ImageFileReader()
            reader.SetImageIO("NiftiImageIO")
            reader.SetFileName(filePath)

        elif filePath[-3:] == "nii":
            reader = sitk.ImageFileReader()
            reader.SetImageIO("NiftiImageIO")
            reader.SetFileName(filePath)

        elif filePath[-4:] == "nrrd":
            reader = sitk.ImageFileReader()
            reader.SetImageIO("NrrdImageIO")
            reader.SetFileName(filePath)

        else:
            reader = sitk.ImageSeriesReader()
            dicom_names = reader.GetGDCMSeriesFileNames(filePath)
            reader.SetFileNames(dicom_names)

        image     = reader.Execute()
        spacing   = image.GetSpacing()
        origin    = image.GetOrigin()  ## bounds
        dimension = image.GetSize()
        extent = (0, dimension[0] - 1, 0, dimension[1] - 1, 0, dimension[2] - 1)

        array = sitk.GetArrayFromImage(image)

        self.itkImage = image
        self.spacing = spacing
        self.origin = origin
        self.dimension = dimension
        self.extent = extent
        self.array = array


    def applyMaskNormalized(self, mask):
        array_mask = np.multiply(mask, self.array)
        origin_peak = np.max(array_mask)
        array_norm_full = self.array/origin_peak
        array_norm_mask = array_mask/origin_peak

        self.array_mask = array_mask
        self.array_norm_full = array_norm_full
        self.array_norm_mask = array_norm_mask
        self.peak_in_brain = origin_peak

    def applyMaskNormalized_onemore(self, mask):
        array_mask = np.multiply(mask, self.array_mask)
        origin_peak = np.max(array_mask)
        array_norm_full = self.array/origin_peak
        array_norm_mask = array_mask/origin_peak

        self.array_mask = array_mask
        self.array_norm_full = array_norm_full
        self.array_norm_mask = array_norm_mask
        self.peak_in_brain = origin_peak

    def makeVTKVolume(self, inputArray):

        if inputArray ==0:
            vtk_pressure = ns.numpy_to_vtk(self.array.flatten())
        else:
            vtk_pressure = ns.numpy_to_vtk(inputArray.flatten())

        image = vtk.vtkImageData()
        image.SetDimensions(l2n(self.dimension) - 1)
        image.SetExtent(self.extent)
        image.SetSpacing(self.spacing)
        image.SetOrigin(self.origin[0], self.origin[1], self.origin[2])
        image.GetPointData().SetScalars(vtk_pressure)

        volumeMapper = vtk.vtkGPUVolumeRayCastMapper()
        volumeMapper.SetInputData(image)

        opacityTransfer = vtk.vtkPiecewiseFunction()
        opacityTransfer.AddPoint(0.05, 0)
        #opacityTransfer.AddPoint(0.05, 0)
        opacityTransfer.AddPoint(0.15, 0.2)
        opacityTransfer.AddPoint(0.5, 0.5)
        opacityTransfer.AddPoint(0.8, 0.8)
        opacityTransfer.AddPoint(0.95, 1)

        ctf = vtk.vtkColorTransferFunction()
        ctf.AddRGBPoint(0.3, 0.1, 0.1, 1.0)
        ctf.AddRGBPoint(0.5, 0.2, 1.0, 0.2)
        ctf.AddRGBPoint(0.65, 1.0, 1.0, 0)
        ctf.AddRGBPoint(0.8, 1.0, 0.5, 0)
        ctf.AddRGBPoint(0.95, 1.0, 0.0, 0.0)

        volumeProperty = vtk.vtkVolumeProperty()
        volumeProperty.SetColor(ctf)
        volumeProperty.SetScalarOpacity(opacityTransfer)
        volumeProperty.SetScalarOpacityUnitDistance(10)

        volume = vtk.vtkVolume()
        volume.SetMapper(volumeMapper)
        volume.SetProperty(volumeProperty)

        return image, volume

    def makeFWHM(self, contor_percent=0.5, set_color=[1,0.3,0]):

        vtkImage = ns.numpy_to_vtk(self.array_norm_mask.flatten())
        image = vtk.vtkImageData()
        image.SetDimensions(l2n(self.dimension) - 1)
        image.SetExtent(self.extent)
        image.SetSpacing(self.spacing)
        image.SetOrigin(self.origin[0], self.origin[1], self.origin[2])
        image.GetPointData().SetScalars(vtkImage)

        contour = vtk.vtkContourFilter()
        contour.SetInputData(image)
        contour.ComputeNormalsOn()
        contour.ComputeScalarsOn()
        contour.ComputeGradientsOn()
        contour.SetValue(0, contor_percent)
        contour.Update()

        transform = vtk.vtkTransform()
        transform.Scale(1, 1, 1)

        transform_filter = vtk.vtkTransformFilter()
        transform_filter.SetInputConnection(contour.GetOutputPort())
        transform_filter.SetTransform(transform)
        transform_filter.Update()

        FWHM_poly = transform_filter.GetOutput()

        center = vtk.vtkCenterOfMass()
        center.SetInputData(FWHM_poly)
        center.Update()

        centroid_FWHM = center.GetCenter()

        mass = vtk.vtkMassProperties()
        mass.SetInputData(FWHM_poly)
        mass.Update()

        volume_FWHM = mass.GetVolume()

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(transform_filter.GetOutputPort())
        mapper.ScalarVisibilityOff()

        actor = vtk.vtkLODActor()
        actor.SetNumberOfCloudPoints(100)
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(set_color)
        actor.GetProperty().SetOpacity(0.7)

        return actor, centroid_FWHM, volume_FWHM

    def segmentationMask(self, dilate, threshold = 250):

        padding_level = 20
        ret, img_th = cv2.threshold(self.array, threshold, 1, 0)

        # remain largest region
        label_im, nb_labels = ndimage.label(img_th)
        sizes = ndimage.sum(img_th, label_im, range(nb_labels + 1))
        mask = sizes == np.max(sizes)
        label_img = mask[label_im]

        # seg_arr = np.multiply(arr, label_img)

        if threshold == -100:
            kernel = np.ones((3, 3), np.uint8)
            open_volume = cv2.morphologyEx(np.array(label_img).astype(np.uint8), cv2.MORPH_OPEN, kernel)

            label_im, nb_labels = ndimage.label(open_volume)
            sizes = ndimage.sum(img_th, label_im, range(nb_labels + 1))
            mask = sizes == np.max(sizes)
            label_img = mask[label_im]
            open_volume = np.array(label_img).astype(np.uint8)

            closed_volume = cv2.morphologyEx(open_volume, cv2.MORPH_CLOSE, kernel)
            # closed_volume = ndimage.binary_erosion(closed_volume)
            #         temp =  closed_volume
            temp = np.zeros_like(label_img)

            for idx in range(closed_volume.shape[0]):
                closed_slice = closed_volume[idx, :, :]
                mask_slice = ndimage.morphology.binary_fill_holes(closed_slice)
                temp[idx, :, :] = mask_slice
            #   mpimg.imsave(os.path.join('E:/2020/KHM_synt/seg', str(threshold)+'_'+str(idx)+'.png'), mask_slice, vmin=-1, vmax=1, cmap='gray')


        if threshold >= 250:
            kernel = np.ones((3, 3), np.uint8)
            open_volume = cv2.morphologyEx(np.array(label_img).astype(np.uint8), cv2.MORPH_OPEN, kernel)

            label_im, nb_labels = ndimage.label(open_volume)
            sizes = ndimage.sum(img_th, label_im, range(nb_labels + 1))
            mask = sizes == np.max(sizes)
            label_img = mask[label_im]
            open_volume = np.array(label_img).astype(np.uint8)

            closed_volume = cv2.morphologyEx(open_volume, cv2.MORPH_CLOSE, kernel)
            temp = closed_volume
            skull_volume = np.array(temp)

        if dilate == True:
            kernel = np.ones((2,2), np.uint8)
            skull_volume = cv2.dilate(np.array(skull_volume).astype(np.uint8),kernel, iterations=10)
        else:
            skull_volume = np.multiply(skull_volume, self.array)
            skull_volume[skull_volume>1] = 1
            skull_volume[skull_volume<1] = 0
        # extract brain region only
        # mask = np.copy(skull_volume)
        #
        # mask[:padding_level, :, :] = 1
        # mask[-padding_level:, :, :] = 1
        # mask[:, :padding_level, :] = 1
        # mask[:, -padding_level:, :] = 1
        # mask[:, :, :padding_level] = 1
        # mask[:, :, -padding_level:] = 1
        # mask = mask - 1
        # mask = -mask
        #
        # brain_mask = np.zeros_like(mask)
        #
        # for i in range(brain_mask.shape[1]):
        #     ret, makers = cv2.connectedComponents(mask[:, i, :])
        #     brain_mask[:, i, :] = makers
        #
        # brain_mask[brain_mask <= 1] = 0
        # brain_mask[brain_mask > 0] = 1
        brain_mask = 0
        return skull_volume, brain_mask

    def makeITK(self, array, path):

        #array = array.transpose([1, 0, 2])
        result_itk = sitk.GetImageFromArray(array, sitk.sitkInt8)
        result_itk.SetSpacing(self.spacing)
        result_itk.SetOrigin(self.origin)
        if path:
            writer = sitk.ImageFileWriter()
            writer.SetFileName(path)
            writer.Execute(result_itk)

        return result_itk

    def saveITK(self, itkimage, path):

        if path:
            writer = sitk.ImageFileWriter()
            writer.SetFileName(path)
            writer.Execute(itkimage)


    def mask2000(self, path):

        itkImage = self.itkImage
        itkImage = itkImage*2000

        writer = sitk.ImageFileWriter()
        writer.SetFileName(path)
        writer.Execute(itkImage)

        return itkImage


    def figure_2D(savename, data, vmin, vmax):
        shape = data.shape
        max_idx = np.where(data == np.max(data))
        fig = plt.figure(figsize=(10, 10))
        gs = fig.add_gridspec(1, 1)
        ax0 = fig.add_subplot(gs[0])
        ax0.imshow(np.rot90(data[:, int(shape[1]/2), :]), cmap='jet', vmin=vmin, vmax=vmax)
        #ax0.scatter(max_idx[0][0], shape[0] - max_idx[2][0], c='white', marker='x', s=400)
        #ax0.scatter(shape[0]/2 - max_idx[2][0], c='white', marker='x', s=400)

        fig.savefig(savename, transparent=True)
        return max_idx

    def figure_2D_compare(savename, real, syn):
        shape = real.shape
        real_max_idx = np.where(real == np.max(real))
        syn_max_idx = np.where(syn == np.max(syn))

        fig = plt.figure(figsize=(50, 50))
        gs = fig.add_gridspec(2, 2)
        ax0 = fig.add_subplot(gs[0])
        ax0.imshow(np.rot90(real[:, real_max_idx[1][0], :]), cmap='jet', interpolation='nearest', vmin=0, vmax=np.max(real)+1)
        ax0.scatter(real_max_idx[0][0], shape[0] - real_max_idx[2][0], c='white', marker='x', s=1000)
        ax0.axis('off')

        ax1 = fig.add_subplot(gs[1])
        ax1.imshow(np.rot90(syn[:, syn_max_idx[1][0], :]), cmap='jet', interpolation='nearest', vmin=0, vmax=np.max(real)+1)
        ax1.scatter(syn_max_idx[0][0], shape[0] - syn_max_idx[2][0], c='white', marker='x', s=1000)
        ax1.axis('off')

        ax2 = fig.add_subplot(gs[2])
        ax2.imshow(np.rot90(real[:, :, real_max_idx[2][0]]), cmap='jet', interpolation='nearest', vmin=0, vmax=np.max(real)+1)
        ax2.scatter(real_max_idx[0][0], shape[0] - real_max_idx[2][0], c='white', marker='x', s=1000)
        ax2.axis('off')

        ax3 = fig.add_subplot(gs[3])
        ax3.imshow(np.rot90(syn[:, :, syn_max_idx[2][0]]), cmap='jet', interpolation='nearest', vmin=0, vmax=np.max(real)+1)
        ax3.scatter(syn_max_idx[0][0], shape[0] - syn_max_idx[1][0], c='white', marker='x', s=1000)
        ax3.axis('off')

        fig.savefig(savename, transparent=True)

        return real_max_idx, syn_max_idx


    def figure_target_2D(savename, data, color_type, vmin, vmax):
        shape = data.shape
        fig = plt.figure(figsize=(10, 10), facecolor='white')

        plt.axis('off')
        plt.imshow(np.fliplr(np.flip(np.squeeze(data[:, :, np.int(shape[1] / 2)]))), cmap=color_type,
                   interpolation='hanning', vmin=vmin, vmax=vmax)
        # plt.colorbar()
        plt.scatter(np.int(shape[1] / 2), np.int(shape[1] / 2), c='white', marker='x', s=400)
        fig.savefig(savename + '.png', transparent=True)
        
        

    def read_1D(self, start, end):
        
        start = l2n(start)
        end = l2n(end)
        
        dir = (end - start)
        unit_dir = dir/np.linalg.norm(dir)

        sorting_len = np.linalg.norm(dir)
        dist = np.linspace(0,sorting_len,np.int(sorting_len*2))

        dist = np.matlib.repmat(dist, 3,1).T
        vectors  = np.matlib.repmat(unit_dir, np.int(sorting_len*2), 1)

        points = start+np.multiply(vectors, dist)
                
        
        axial_result = np.zeros(len(points))

        for i in range(len(points)):
            idx = self.itkImage.TransformPhysicalPointToIndex(points[i,:])
            if idx[0]>=self.dimension[0] or idx[1]>=self.dimension[1] or idx[2]>=self.dimension[2] :
                axial_result[i] = 0
            else:
                axial_result[i] = self.itkImage[idx]
            
        return axial_result
